@@ -27,10 +27,13 @@
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
 #include <unistd.h>
 #include <ctype.h>
 #include <time.h>
 
+#include <glib/gprintf.h>
+
 #include "exmap.h"
 
 #define EXIT_KERNEL_ERROR  -1
@@ -59,25 +62,65 @@
 #ifdef INTERACTIVE
 extern int rows;
 extern int columns;
-
-#define myprintf(args...)                                   \
-do {                                                        \
-    if (of == stdout && global_lines >= rows - 1)           \
-      {                                                     \
-        global_lines = 0;                                   \
-        fprintf (of, "  --- Press Enter for more ---");     \
-        fgetc (stdin);                                      \
-      }                                                     \
-                                                            \
-    fprintf (args);                                         \
-    global_lines++;                                         \
-}while (0)
-
-#else
-#define myprintf(args...) fprintf (args)
 #endif
 
-static gint global_lines = 0;
+static void
+myprintf (data_t *d, const char *fmt, ...)
+{
+  va_list args;
+
+  if (! (d && d->of && fmt))
+    return;
+
+#ifdef INTERACTIVE
+  if (d->args.interactive && d->global_lines >= rows - 1)
+    {
+      d->global_lines = 0;
+      fprintf (d->of, "  --- Press Enter for more ---");
+      fgetc (stdin);
+    }
+#endif
+  
+  va_start (args, fmt);
+  g_vfprintf (d->of, fmt, args);
+  va_end (args);
+
+  d->global_lines++;
+}
+
+guint
+print_paged (data_t * d, char * text)
+{
+  if (!d || !text)
+    return 0;
+
+  d->global_lines = 0;
+  
+  char * t = text;
+  char * nl = strchr (t, '\n');
+  guint count = 0;
+  
+  while (nl)
+    {
+      ++count;
+      *nl = 0;
+
+      myprintf (d, "%s\n", t);
+      t = nl + 1;
+      
+      nl = strchr (t, '\n');
+    }
+  
+
+  if (*t)
+    {
+      myprintf (d, "%s\n", t);
+      ++count;
+    }
+
+  return count;
+}
+
 
 address_t
 page_start (address_t addr)
@@ -430,7 +473,7 @@ process_files (gpointer key, gpointer value, gpointer data)
 }
 
 static guint
-print_symbol_info (FILE * of, symbol_t * s, data_t * d)
+print_symbol_info (symbol_t * s, data_t * d)
 {
   if (!s || !s->m.vm)
     return 0;
@@ -439,7 +482,7 @@ print_symbol_info (FILE * of, symbol_t * s, data_t * d)
   
   if (d->args.format == FORMAT_HUMAN)
     {
-      myprintf (of, "      %-26s: VM%8u %s, M%8u %s, S%8u %s\n",
+      myprintf (d, "      %-26s: VM%8u %s, M%8u %s, S%8u %s\n",
                s->name,
                SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
                SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
@@ -449,7 +492,7 @@ print_symbol_info (FILE * of, symbol_t * s, data_t * d)
     }
   else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      myprintf (of, ",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+      myprintf (d, ",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
                s->name, 
                s->m.vm,
                s->m.mapped,
@@ -463,7 +506,7 @@ print_symbol_info (FILE * of, symbol_t * s, data_t * d)
 
 /* TODO refactor these print fuctions to share code */
 static guint
-print_section_info (FILE * of, section_t * s, data_t * d)
+print_section_info (section_t * s, data_t * d)
 {
   if (!s)
     return 0;
@@ -474,7 +517,7 @@ print_section_info (FILE * of, section_t * s, data_t * d)
     {
       if (d->args.format == FORMAT_HUMAN)
         {
-          myprintf (of,"   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
+          myprintf (d,"   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
                    s->name,
                    SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
                    SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
@@ -485,7 +528,7 @@ print_section_info (FILE * of, section_t * s, data_t * d)
       else if (d->args.format == FORMAT_CSV ||
                d->args.format == FORMAT_CSV_SUMS)
         {
-          myprintf (of,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+          myprintf (d,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
                    s->name, 
                    s->m.vm,
                    s->m.mapped,
@@ -498,13 +541,13 @@ print_section_info (FILE * of, section_t * s, data_t * d)
     {
       if (d->args.format == FORMAT_HUMAN)
         {
-          myprintf (of,"   %-29s:\n", s->name);
+          myprintf (d,"   %-29s:\n", s->name);
           lines++;
         }
       else if (d->args.format == FORMAT_CSV ||
                d->args.format == FORMAT_CSV_SUMS)
         {
-          myprintf (of,",,,,,,,,,,,,\"%s\",0,0,0\n", s->name);
+          myprintf (d,",,,,,,,,,,,,\"%s\",0,0,0\n", s->name);
           lines++;
         }
     }
@@ -513,7 +556,7 @@ print_section_info (FILE * of, section_t * s, data_t * d)
   while (l)
     {
       symbol_t * sbl = l->data;
-      lines += print_symbol_info (of, sbl, d);
+      lines += print_symbol_info (sbl, d);
       l = l->next;
     }
   
@@ -521,7 +564,7 @@ print_section_info (FILE * of, section_t * s, data_t * d)
 }
 
 static guint
-print_file_info (FILE * of, file_t * f, data_t * d)
+print_file_info (file_t * f, data_t * d)
 {
   if (!f)
     return 0;
@@ -533,17 +576,17 @@ print_file_info (FILE * of, file_t * f, data_t * d)
     {
       if (d->args.detail & DETAIL_ELF_INFO)
         {
-          myprintf (of,"  ----------------------------------"
+          myprintf (d,"  ----------------------------------"
                    "---------------------------------------\n");
           lines++;
         }
 
-      myprintf (of,"  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
+      myprintf (d,"  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
                cmd,
                SCALE_V(f->m.vm),     SCALE_U(f->m.vm),
                SCALE_V(f->m.mapped), SCALE_U(f->m.mapped),
                SCALE_V(f->m.sole),   SCALE_U(f->m.sole));
-      global_lines--;
+      d->global_lines--;
       
 #if 0
       elf_print_dbg_info (f->elf);
@@ -551,19 +594,19 @@ print_file_info (FILE * of, file_t * f, data_t * d)
       if (d->args.detail & DETAIL_ELF_INFO)
         {
           lines += 2;
-          myprintf (of,"\n  ----------------------------------"
+          myprintf (d,"\n  ----------------------------------"
                    "---------------------------------------\n");
-          global_lines++;
+          d->global_lines++;
         }
       else
         {
           lines++;
-          myprintf (of,"\n");
+          myprintf (d,"\n");
         }
     }
   else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      myprintf (of,",,,,,,,,\"%s\",%u,%u,%u\n",
+      myprintf (d,",,,,,,,,\"%s\",%u,%u,%u\n",
                f->m.cmd, 
                f->m.vm,
                f->m.mapped,
@@ -580,7 +623,7 @@ print_file_info (FILE * of, file_t * f, data_t * d)
       while (l)
         {
           section_t * s= (section_t*)l->data;
-          lines += print_section_info (of, s, d);
+          lines += print_section_info (s, d);
           l = l->next;
         }
     }
@@ -1042,7 +1085,7 @@ sort_pid_list (gconstpointer a, gconstpointer b, gpointer data)
  * file_info indicates if per-file information is to be printed as well
  */
 static guint
-print_results (FILE * of, memory_t * mem_data, data_t * d)
+print_results (memory_t * mem_data, data_t * d)
 {
   guint lines = 0;
 
@@ -1054,29 +1097,29 @@ print_results (FILE * of, memory_t * mem_data, data_t * d)
 
   if (d->args.format == FORMAT_HUMAN)
     {
-      myprintf (of,"Process %d", mem_data->pid);
-      global_lines--;
+      myprintf (d,"Process %d", mem_data->pid);
+      d->global_lines--;
       
       if (mem_data->cmd)
-        myprintf (of," [%s]\n", mem_data->cmd);
+        myprintf (d," [%s]\n", mem_data->cmd);
       else
-        myprintf (of,"\n");
+        myprintf (d,"\n");
   
-      myprintf (of," Virtual memory  : %9u %s\n",
+      myprintf (d," Virtual memory  : %9u %s\n",
                SCALE_V(mem_data->vm), SCALE_U(mem_data->vm));
-      myprintf (of," Effective VM    : %9u %s\n",
+      myprintf (d," Effective VM    : %9u %s\n",
                SCALE_V(mem_data->evm), SCALE_U(mem_data->evm));
-      myprintf (of," Heap            : %9u %s\n",
+      myprintf (d," Heap            : %9u %s\n",
                SCALE_V(mem_data->heap), SCALE_U(mem_data->heap));
-      myprintf (of," Stack           : %9u %s\n",
+      myprintf (d," Stack           : %9u %s\n",
                SCALE_V(mem_data->stack), SCALE_U(mem_data->stack));
-      myprintf (of," vdso            : %9u %s\n",
+      myprintf (d," vdso            : %9u %s\n",
                SCALE_V(mem_data->vdso), SCALE_U(mem_data->vdso));
-      myprintf (of," Mapped          : %9u %s\n",
+      myprintf (d," Mapped          : %9u %s\n",
                SCALE_V(mem_data->mapped), SCALE_U(mem_data->mapped));
-      myprintf (of," Effective mapped: %9u %s\n",
+      myprintf (d," Effective mapped: %9u %s\n",
                SCALE_V(mem_data->emapped), SCALE_U(mem_data->emapped));
-      myprintf (of," Sole use        : %9u %s\n",
+      myprintf (d," Sole use        : %9u %s\n",
                SCALE_V(mem_data->sole), SCALE_U(mem_data->sole));
 
       lines += 9;
@@ -1084,15 +1127,15 @@ print_results (FILE * of, memory_t * mem_data, data_t * d)
       if (d->args.detail & DETAIL_FILE_INFO)
         {
           lines += 2;
-          global_lines++;
-          myprintf (of,"\n Per file memory use\n");
+          d->global_lines++;
+          myprintf (d,"\n Per file memory use\n");
         }
     }
   else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
       lines++;
       
-      myprintf (of,"%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
+      myprintf (d,"%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
                mem_data->pid,
                mem_data->cmd ? mem_data->cmd : "???",
                mem_data->vm,
@@ -1116,7 +1159,7 @@ print_results (FILE * of, memory_t * mem_data, data_t * d)
                                      GINT_TO_POINTER (d->args.sort));
           while (l)
             {
-              lines += print_file_info (of, (file_t*)l->data, d);
+              lines += print_file_info ((file_t*)l->data, d);
               l = l->next;
             }
 
@@ -1128,12 +1171,12 @@ print_results (FILE * of, memory_t * mem_data, data_t * d)
     {
       if (!d->args.interactive)
         {
-          myprintf (of,"\n========================================"
+          myprintf (d,"\n========================================"
                        "=======================================\n\n");
-          global_lines += 2;
+          d->global_lines += 2;
         }
       else
-        myprintf (of, "\n");
+        myprintf (d, "\n");
     }
 
   return lines;
@@ -1434,9 +1477,8 @@ main_print (data_t * d)
   guint        lines = 0;
   GList      * l;
   pid_t      * pid;
-  FILE       * of = d->of;
 
-  global_lines = 0;
+  d->global_lines = 0;
   
   if (!d->args.interactive)
     {
@@ -1450,31 +1492,31 @@ main_print (data_t * d)
       strftime (date_stamp, sizeof (date_stamp),
                 "%a %Y %m %d %H:%M:%S", &t);
                       
-      fprintf (of,"--- Exmap log from %s ---\n\n", date_stamp);
+      fprintf (d->of,"--- Exmap log from %s ---\n\n", date_stamp);
     }
   
   /* output header */
   if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      myprintf (of,
+      myprintf (d,
                "pid,command,VM,effective VM,heap,stack,vdso,mapped,"
                "emapped,sole");
-      global_lines--;
+      d->global_lines--;
       
       if (d->args.detail & DETAIL_FILE_INFO)
         {
-          myprintf (of,",file,file-VM,file-mapped,file-solefile");
+          myprintf (d,",file,file-VM,file-mapped,file-solefile");
 
           if (d->args.detail & DETAIL_ELF_INFO)
-            myprintf (of,
+            myprintf (d,
                      ",section,section-VM,section-mapped,section-sole\n");
           else
-            myprintf (of,"\n");
+            myprintf (d,"\n");
 
-          global_lines--;
+          d->global_lines--;
         }
       else
-        myprintf (of,"\n");
+        myprintf (d,"\n");
     }
 
   /* output info for individula pids */
@@ -1483,7 +1525,7 @@ main_print (data_t * d)
       l = d->mem_list;
       while (l)
         {
-          lines += print_results (d->of, (memory_t*)l->data, d);
+          lines += print_results ((memory_t*)l->data, d);
           l = l->next;
         }
     }
@@ -1498,7 +1540,7 @@ main_print (data_t * d)
                                       GINT_TO_POINTER (*pid), find_pid);
 
               if (l)
-                lines += print_results (d->of, (memory_t*)l->data, d);
+                lines += print_results ((memory_t*)l->data, d);
               pid++;
             }
         }
