@@ -154,9 +154,11 @@ typedef struct
   FILE *       of;
   GList      * mem_list;
   args_t       args;
+  gboolean     global_lines;
 }data_t;
 
 
+guint print_paged (data_t * d, char * text);
 void main_load (data_t * d);
 void main_reset (data_t * d);
 
