@@ -1,5 +1,5 @@
 /* -*- mode:C; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
-/* 
+/*
  * Copyright (C) 2006 OpenedHand Ltd.
  *
  * Author: Tomas Frydrych <tf@o-hand.com>
@@ -27,6 +27,7 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <pwd.h>
 
 #include <readline/readline.h>
@@ -44,6 +45,7 @@ enum
     EMSG_PID_NOT_SELECTED,
     EMSG_NO_PROCESS_SELECTED,
     EMSG_FOPEN_ERROR,
+    EMSG_UNKNOWN_ERROR,
 
     /* MUST BE LAST, HAVE NO CORRESPONDING STRING */
     EMSG_ERROR_NO_MESSAGE, /* for errors for which msg was already printed */
@@ -57,6 +59,7 @@ static char * emsg [] =
     "  Process %d is not selected\n",
     "  No process selected\n",
     "  Could not open file [%s]\n",
+    "  Unknown error occured\n"
   };
 
 static gchar *
@@ -110,13 +113,13 @@ cmd_to_pid (data_t * d, const char * cmd)
               c++;
               name = c;
             }
-      
+
           if (!strncmp (name, cmd, strlen (cmd)))
             {
               result = g_list_prepend (result, GINT_TO_POINTER (m->pid));
             }
         }
-      
+
       l = l->next;
     }
 
@@ -142,7 +145,7 @@ pid_list_to_array (GList * l, pid_t * old_pids, int * count)
       pids = g_malloc ((len + 1) * sizeof (pid_t));
       *count = 0;
     }
-  
+
   for (i = 0; i < len; ++i)
     {
       pid_t pid = GPOINTER_TO_INT (l->data);
@@ -165,7 +168,7 @@ interactive_print (data_t * d, const char * p)
     {
       GList  * l;
       pid_t    pid;
-      
+
       if (!strcmp (p, "all"))
         {
           d->args.all_pids = TRUE;
@@ -177,7 +180,7 @@ interactive_print (data_t * d, const char * p)
         {
           pid_t  * old_pids  = d->args.pids_arg;
           guint    old_count = d->args.pids_count;
-          
+
           if ((pid = atoi (p)))
             {
               d->args.pids_arg = g_malloc0(2 * sizeof(pid_t));
@@ -195,7 +198,7 @@ interactive_print (data_t * d, const char * p)
             {
               return EMSG_INVALID_PARAMETER;
             }
-          
+
           main_print (d);
 
           g_free (d->args.pids_arg);
@@ -216,12 +219,16 @@ interactive_add (data_t *d, const char * p)
 {
   GList * l;
   pid_t   pid;
-  
-  if (*p && (pid = atoi (p)))
+
+  if (!strcmp (p, "all"))
+    {
+      d->args.all_pids = TRUE;
+    }
+  else if (*p && (pid = atoi (p)))
     {
       d->args.pids_arg = g_realloc(d->args.pids_arg,
                                    (d->args.pids_count + 2)*sizeof(pid_t));
-          
+
       d->args.pids_arg[d->args.pids_count] = pid;
       d->args.pids_count++;
       d->args.pids_arg[d->args.pids_count] = 0;
@@ -234,19 +241,21 @@ interactive_add (data_t *d, const char * p)
   else
     return EMSG_INVALID_PARAMETER;
 
-  return EMSG_SUCCESS;  
+  return EMSG_SUCCESS;
 }
 
 static guint
 interactive_rm (data_t * d, const char * p)
 {
   pid_t pid;
-  
-  if (*p && (pid = atoi (p)))
+
+  if (!strcmp (p, "all"))
+    d->args.all_pids = FALSE;
+  else if (*p && (pid = atoi (p)))
     {
       pid_t * pp = d->args.pids_arg;
       gboolean found = FALSE;
-              
+
       while (*pp)
         {
           if (*pp == pid)
@@ -254,7 +263,7 @@ interactive_rm (data_t * d, const char * p)
               found = TRUE;
               break;
             }
-                  
+
           pp++;
         }
 
@@ -265,6 +274,8 @@ interactive_rm (data_t * d, const char * p)
               *pp = *(pp + 1);
               pp++;
             }
+
+          d->args.pids_count--;
         }
       else
         {
@@ -281,23 +292,28 @@ interactive_rm (data_t * d, const char * p)
 static guint
 interactive_ls (data_t *d)
 {
-  pid_t * p = d->args.pids_arg;
-  if (p && *p)
-    {
-      while (*p)
-        {
-          g_print ("%d ", *p);
-          p++;
-        }
-
-      g_print ("\n");
-    }
+  if (d->args.all_pids)
+    g_print ("'all'\n");
   else
-    return EMSG_NO_PROCESS_SELECTED;
+    {
+      pid_t * p = d->args.pids_arg;
+      if (p && *p)
+        {
+          while (*p)
+            {
+              g_print ("%d ", *p);
+              p++;
+            }
+
+          g_print ("\n");
+        }
+      else
+        return EMSG_NO_PROCESS_SELECTED;
+    }
 
   return EMSG_SUCCESS;
 }
-     
+
 static guint
 interactive_save (data_t * d, const char * file)
 {
@@ -313,9 +329,9 @@ interactive_save (data_t * d, const char * file)
   /* step out of interactive mode ... */
   d->of = of;
   d->args.interactive = FALSE;
-  
+
   main_print (d);
-  
+
   d->of = old_of;
   d->args.interactive = TRUE;
 
@@ -329,14 +345,18 @@ interactive_help ()
 {
   g_print (
       "Available commands:\n"
-      "  add pid|command             : select process for examination\n"
+      "  add 'all'|pid|command       : select process for examination\n"
       "  clear                       : deselect all selected processes\n"
       "  detail 'none'|'file'|'elf'  : set level of detail to show\n"
       "  ls                          : list selected processes\n"
       "  print ['all'|pid|command]   : print data for selected/specified processes\n"
+      "  ps [options]                : run system ps command\n"
       "  reload                      : recreate system snapshot\n"
-      "  rm pid                      : deselect process with pid\n"
+      "  rm pid|'all'                : deselect process with pid, or clear 'all' flag\n"
       "  save file                   : save data for selected process to file\n"
+      "  shell commands              : run shell commands\n"
+      "  sort asc|desc|vm|sole|heap|stack|mapped|emapped\n"
+      "                              : sort results by\n"
       "  threshold                   : set minimum VM for processes printed\n"
       "  ?, help                     : help\n"
       "  quit                        : quit\n");
@@ -344,6 +364,51 @@ interactive_help ()
   return EMSG_SUCCESS;
 }
 
+static guint
+interactive_spawn (data_t *d, const char * cmd, const char * params)
+{
+  GError * error = NULL;
+  gchar  * out = NULL;
+  gchar  * err = NULL;
+  gint     status;
+  guint    retval = EMSG_SUCCESS;
+  gchar  * cmdline = NULL;
+
+  cmdline = g_strconcat (cmd, " ", params, NULL);
+
+  d->global_lines = 0;
+
+  if (g_spawn_command_line_sync (cmdline, &out, &err, &status, &error))
+    {
+      if (out)
+        {
+          print_paged (d, out);
+          g_free (out);
+        }
+
+      if (WEXITSTATUS (status))
+        {
+          if (err)
+            {
+              print_paged (d, err);
+              g_free (err);
+            }
+
+          retval = EMSG_UNKNOWN_ERROR;
+        }
+    }
+  else
+    {
+      print_paged (d, error->message);
+      g_error_free (error);
+      retval = EMSG_UNKNOWN_ERROR;
+    }
+
+  g_free (cmdline);
+
+  return EMSG_SUCCESS;
+}
+
 static gboolean
 interactive_parse_line (data_t *d, char * line)
 {
@@ -353,12 +418,12 @@ interactive_parse_line (data_t *d, char * line)
   guint      retval = EMSG_SUCCESS;
   gchar    * cmd;
   size_t     cmd_len;
-  
+
   if (!line)
     return FALSE;
 
   cmd = g_strdup (line);
-  
+
   while (isspace (*cmd))
     cmd++;
 
@@ -375,7 +440,7 @@ interactive_parse_line (data_t *d, char * line)
     }
 
   cmd_len = strlen (cmd);
-  
+
   switch (c)
     {
     case 'a':
@@ -388,6 +453,7 @@ interactive_parse_line (data_t *d, char * line)
     case 'c':
       if (!strncmp (cmd, "clear", cmd_len))
         {
+          d->args.all_pids = FALSE;
           g_free (d->args.pids_arg);
           d->args.pids_arg = NULL;
           d->args.pids_count = 0;
@@ -395,7 +461,7 @@ interactive_parse_line (data_t *d, char * line)
       else
         retval = EMSG_UNKNOWN_COMMAND;
       break;
-      
+
     case 'd':
       if (!strncmp (cmd, "detail", cmd_len))
         {
@@ -414,14 +480,14 @@ interactive_parse_line (data_t *d, char * line)
       else
         retval = EMSG_UNKNOWN_COMMAND;
       break;
-      
+
     case 'h':
       if (!strncmp (cmd, "help", cmd_len))
         retval = interactive_help ();
       else
         retval = interactive_help ();
       break;
-      
+
     case '?':
       retval = interactive_help ();
       break;
@@ -432,9 +498,11 @@ interactive_parse_line (data_t *d, char * line)
       else
         retval = EMSG_UNKNOWN_COMMAND;
       break;
-      
+
     case 'p':
-      if (!strncmp (cmd, "print", cmd_len))
+      if (!strcmp (cmd, "ps"))
+        retval = interactive_spawn (d, "/bin/ps", p);
+      else if (!strncmp (cmd, "print", cmd_len))
         retval = interactive_print (d, p);
       else
         retval = EMSG_UNKNOWN_COMMAND;
@@ -457,9 +525,36 @@ interactive_parse_line (data_t *d, char * line)
       else
         retval = EMSG_UNKNOWN_COMMAND;
       break;
-      
+
     case 's':
-      if (!strncmp (cmd, "save", cmd_len))
+      if (!strncmp (cmd, "sort", cmd_len))
+        {
+          if (!strcmp (p, "asc"))
+            SORT_TYPE_SET (d->args.sort, SORT_TYPE_ASC);
+          else if (!strcmp (p, "desc"))
+            SORT_TYPE_SET (d->args.sort, SORT_TYPE_DESC);
+          else if (!strcmp (p, "vm"))
+            SORT_BY_SET (d->args.sort, SORT_BY_VM);
+          else if (!strcmp (p, "sole"))
+            SORT_BY_SET (d->args.sort, SORT_BY_SOLE);
+          else if (!strcmp (p, "heap"))
+            SORT_BY_SET (d->args.sort, SORT_BY_HEAP);
+          else if (!strcmp (p, "stack"))
+            SORT_BY_SET (d->args.sort, SORT_BY_STACK);
+          else if (!strcmp (p, "mapped"))
+            SORT_BY_SET (d->args.sort, SORT_BY_MAPPED);
+          else if (!strcmp (p, "emapped"))
+            SORT_BY_SET (d->args.sort, SORT_BY_EMAPPED);
+
+          main_sort (d);
+        }
+      else if (!strncmp (cmd, "shell", cmd_len))
+        {
+          gchar * s = g_strconcat ("\"", p, "\"", NULL);
+          retval = interactive_spawn (d, "/bin/sh -r -c", s);
+          g_free (s);
+        }
+      else if (!strncmp (cmd, "save", cmd_len))
         retval = interactive_save (d, p);
       else
         retval = EMSG_UNKNOWN_COMMAND;
@@ -490,18 +585,19 @@ interactive_parse_line (data_t *d, char * line)
         case EMSG_UNKNOWN_COMMAND:
           g_print (emsg[retval], cmd);
           break;
-          
+
+        case EMSG_UNKNOWN_ERROR:
         case EMSG_ERROR_NO_MESSAGE:
           break;
-          
+
         default:
           g_warning ("Unhandled error value %d", retval);
-          
+
         }
-      
+
       return FALSE;
     }
-  
+
   if (history)
     {
       /* if the line already exists in the history, we remove it
@@ -522,10 +618,10 @@ interactive_parse_line (data_t *d, char * line)
 #endif
             }
         }
-      
+
       add_history (line);
     }
-  
+
   return FALSE;
 }
 
